;;; Copyright (c) 2006 David Lichteblau
;;;
;;; Permission is hereby granted, free of charge, to any person
;;; obtaining a copy of this software and associated documentation files
;;; (the "Software"), to deal in the Software without restriction,
;;; including without limitation the rights to use, copy, modify, merge,
;;; publish, distribute, sublicense, and/or sell copies of the Software,
;;; and to permit persons to whom the Software is furnished to do so,
;;; subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
;;; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
;;; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;;; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;;; SOFTWARE.

(in-package :sb-heapdump)

(defun dump-packages
    (packages pathname
     &rest keys &key if-exists parameters print-statistics initializer)
  (declare (ignore if-exists parameters print-statistics))
  (unless (listp packages)
    (setf packages (list packages)))
  (setf packages
	(mapcar (lambda (p)
		  (or (find-package p) (error "package not found: ~A" p)))
		packages))
  (unless initializer
    (setf initializer #'identity))
  (let ((infos '())
	(find-class-cells '())
	(methods '())
	(fast-methods '()))
    (dolist (package packages)
      (do-symbols (sym package)
	(when (eq (symbol-package sym) package)
	  (setf infos (nconc (infos sym) (infos `(setf ,sym) :function) infos))
	  (let ((cell (gethash sym sb-pcl::*find-class*)))
	    (when cell
	      (push cell find-class-cells)
	      (push sym find-class-cells)
	      (let ((class (sb-pcl::find-class-cell-class cell)))
		(when class
		  (dolist (slot (sb-mop:class-slots class))
		    (dolist (rwb '(sb-pcl::reader
				   sb-pcl::writer
				   sb-pcl::boundp))
		      (setf infos (nconc (infos `(sb-pcl::slot-accessor
						  :global
						  ,(sb-mop:slot-definition-name slot)
						  ,rwb)
						:function)
					 infos))))
		  (dolist (method (sb-mop:specializer-direct-methods class))
		    (let* ((gf (sb-mop:method-generic-function method))
			   (id (function-name-identifier
				(sb-mop:generic-function-name gf))))
		      (let ((fm (sb-pcl::method-fast-function method)))
			(when fm
			  (when
			      ;; FIXME!
			      (eq (car (sb-kernel:%fun-name fm))
				  'sb-pcl::fast-method)
			    (push fm fast-methods)
			    (setf infos (nconc (infos (sb-kernel:%fun-name fm) :function) infos)))))
		      #+(or)
		      (let ((fm (slot-value method 'sb-pcl::function)))
			(when fm
			  (when
			      ;; FIXME!
			      (eq (car (sb-kernel:%fun-name fm))
				  'sb-pcl::fast-method)
			    (push fm fast-methods)
			    (setf infos (nconc (infos (sb-kernel:%fun-name fm) :function) infos)))))
		      (unless (and id (member (symbol-package id) packages))
			(push method methods)
			(push gf methods)))))))))))
    (flet ((update-various-funny-tables (cons)
	     (dolist (package (car cons))
	       (sb-impl::enter-new-nicknames
		package
		(cons (package-name package) (package-nicknames package))))
	     (loop for (sym class . plist) in infos do
		   (loop for (type def) on plist by #'cddr do
			 (setf (sb-int:info class type sym) def)))
	     (loop for (sym cell) on find-class-cells by #'cddr do
		   (setf (gethash sym sb-pcl::*find-class*) cell))
	     (loop for (gf method) on methods by #'cddr do
		   (unless (find method (sb-mop:generic-function-methods gf))
		     (setf (sb-mop:method-generic-function method) nil)
		     (sb-mop:add-method gf method)))
	     (funcall initializer (car cons))))
      (apply #'dump-object
	     (cons packages fast-methods)
	     pathname
	     :force packages
	     :initializer #'update-various-funny-tables
	     keys))))

(defun infos (name &optional class)
  (let ((result '()))
    (maphash (lambda (c class-info)
	       (when (or (null class) (eq c class))
		 (let ((types (sb-c::class-info-types class-info)))
		   (let ((plist
			  (loop
			      for type-info in types
			      for type = (sb-c::type-info-name type-info)
			      for (def hit)
				  := (multiple-value-list
				      (handler-case
					  (sb-int:info c type name)
					;; kludge: there doesn't seem to be a
					;; way to suppress default values, and
					;; some of them throw errors.
					(sb-int:bug ()
					  nil)))
			      when hit
			      append (list type def))))
		     (when plist
		       (push (list* name c plist) result))))))
	     sb-c::*info-classes*)
    result))
